// SPDX-License-Identifier: MIT

syntax = "proto2";

message ipc_desc_entry {
	required uint32		key	= 1;
	required uint32		uid	= 2;
	required uint32		gid	= 3;
	required uint32		cuid	= 4;
	required uint32		cgid	= 5;
	required uint32		mode	= 6;
	required uint32		id	= 7;
}
