// SPDX-License-Identifier: MIT

syntax = "proto2";

enum remap_type {
  LINKED		= 0;
  GHOST 		= 1;
  PROCFS		= 2;
  // Reserved for spfs manager
  // SPFS		= 3;
  // SPFS_LINKED	= 4;
};

message remap_file_path_entry {
	required uint32		orig_id		= 1;
	required uint32		remap_id	= 2;
	optional remap_type	remap_type	= 3;
}
