// SPDX-License-Identifier: MIT

syntax = "proto2";

import "opts.proto";
import "fown.proto";

message pipe_entry {
	required uint32		id		= 1;
	required uint32		pipe_id		= 2;
	required uint32		flags		= 3 [(criu).hex = true];
	required fown_entry	fown		= 4;
}
