// SPDX-License-Identifier: MIT

syntax = "proto2";

import "opts.proto";

message pagemap_head {
	required uint32 pages_id	= 1;
}

message pagemap_entry {
	required uint64 vaddr		= 1 [(criu).hex = true];
	required uint32 nr_pages	= 2;
	optional bool	in_parent	= 3;
	optional uint32	flags		= 4 [(criu).flags = "pmap.flags" ];
}
