#ifndef COMPEL_ELF64_TYPES_H__
#define COMPEL_ELF64_TYPES_H__

#define Elf_Ehdr Elf64_Ehdr
#define Elf_Shdr Elf64_Shdr
#define Elf_Sym	 Elf64_Sym
#define Elf_Rel	 Elf64_Rel
#define Elf_Rela Elf64_Rela

#define ELF_ST_TYPE ELF64_ST_TYPE
#define ELF_ST_BIND ELF64_ST_BIND

#define ELF_R_SYM  ELF64_R_SYM
#define ELF_R_TYPE ELF64_R_TYPE

#endif /* COMPEL_ELF64_TYPES_H__ */
