#ifndef COMPEL_ELF32_TYPES_H__
#define COMPEL_ELF32_TYPES_H__

#define Elf_Ehdr Elf32_Ehdr
#define Elf_Shdr Elf32_Shdr
#define Elf_Sym	 Elf32_Sym
#define Elf_Rel	 Elf32_Rel
#define Elf_Rela Elf32_Rela

#define ELF_ST_TYPE ELF32_ST_TYPE
#define ELF_ST_BIND ELF32_ST_BIND

#define ELF_R_SYM  ELF32_R_SYM
#define ELF_R_TYPE ELF32_R_TYPE

#endif /* COMPEL_ELF32_TYPES_H__ */
